
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Domain file automatically generated by the Tarski FSTRIPS writer
;;; {header_info}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (domain {domain_name})
    (:requirements {requirements})
    (:types
        {types}
    )

    (:constants
        {constants}
    )

    (:predicates
        {predicates}
    )

    (:functions
        {functions}
    )

    {derived}

    {actions}
)