
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Instance file automatically generated by the Tarski FSTRIPS writer
;;; {header_info}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (problem {problem_name})
    (:domain {domain_name})

    (:objects
        {objects}
    )

    (:init
        {init}
    )

    (:goal
        {goal}
    )

    {constraints}
    {domain_bounds}
    {metric}
)

